// $Id: tfreg.cpp 1013 2018-01-04 16:38:27Z ge $
/// \file tfreg.cpp
/// \brief Calculates temperatures using the regression method
/// \author Gerald Weber <gweberbh@gmail.com>
/// $Revision: 1013 $

#include "Options.h"
#include "NeighbourSequence.h"
#include "Regression.h"
#include <unistd.h>
#include "HeterogenousTM.h"
//#include "Actions.h"
#include "FileParser.h"
#include "RegressionModel.h"
#include "CPUtime.h"
namespace gbc
  {
  template<>
  std::valarray<HeterogenousTM<>::value_type> ActionAverageY::action(HeterogenousTM<>& tf,size_t N)
    {
    number_of_basepairs=tf.pSI->BP_reduced_duplex.size();
    tf.retrieve_eigensystem();
    tf.retrieve_or_calculate_matrices();
    return tf.average_y();
    }
  };

using namespace gbc;
using namespace std;

typedef double value_type;

int main(int argc,char *argv[])
  {
  
#ifndef SVNVERSION
#define SVNVERSION "Revision: 865"
#endif
#ifndef LINUXSYSTEM
#define LINUXSYSTEM "UKNOWN"
#endif
#ifndef BUILDTIME
#define BUILDTIME "UKNOWN"
#endif
   
  std::cout << "*** tfreg version=" << std::string(SVNVERSION) << " system=" << std::string(LINUXSYSTEM) << " builddate=" << std::string(BUILDTIME) << std::endl;

 
  Option<string>         output_name("o","output files basename",OptionMap::necessary);
  Option<string>         reg_name("reg","regression parameter file",OptionMap::optional);
  Option<string>         result("res","select which type of results",OptionMap::optional,"regression");
  Option<string>         duplextype("duplextype","type of duplex, e.g. DNA or RNA",OptionMap::optional,"DNA");
  Option<deque<string> > par_file("par","list of parameter file",OptionMap::necessary);
  Option<string>         expand("expand","expand on this base",OptionMap::optional,"CG_CG");
  Option<string>         matrixbase("matrix","location of the matrices",OptionMap::optional);
  Option<string>         datafile("data","parsed data file",OptionMap::optional);
  Option<string>         debug_file("debugfile","file to be debugged",OptionMap::optional);
  Option<string>         sequence("seq","sequence (5' to 3')",OptionMap::optional);
  Option<string>         csequence("cseq","complementary sequence (3' to 5')",OptionMap::optional);
  Option<int>            cutoff("cutoff","number of eigenvalues at which to cutoff a summation",OptionMap::optional,0);
  Option<bool>           periodic_bc("pbc","periodic boundary condition",OptionMap::optional,false);
  Option<bool>           debug("debug","select debuging",OptionMap::optional,false);
  Option<std::string>    model("m","model",OptionMap::optional,"pb");
  Option<double>         temperature("t","temperature in Kelvin",OptionMap::optional,370.0);
  Option<Range<double> > temperature_range("tr","temperature range for looping in Kelvin",OptionMap::optional);
  Option<Range<double> > integration("int","integration range",OptionMap::optional);
  Option<double>         exp_err("ee","experimental error",OptionMap::optional,0.0);
  Option<int>            ran_seed("rs","random_seed",OptionMap::optional,1);
  Option<int>            min_length_regression("mlr","minimum duplex length to be included in regression",OptionMap::optional,0);
  Option<int>            min_a_regression("mar","minimum amount of data points for a given N in regression",OptionMap::optional,3);
  Option<int>            prediction_method("pm","prediction method",OptionMap::optional,2);
  Option<double>         salt("salt","salt concentration (mM)", OptionMap::optional,69);
  Option<map<string,string> > dict("dict","Additions to nucleotide dictionary",OptionMap::optional);
  Option<string>         terminal_char("terminal","which character should represent the terminal of a sequence",OptionMap::optional,std::string("_"));

  OptionMap::arguments(argc,argv);
  OptionMap::scan_arguments();
  if (result.Value=="nncheck")
    {
    output_name.turn_optional();
    par_file.turn_optional();
    if (!sequence.provided()) datafile.turn_necessary();
    }
    
  if (result.Value == "zyfy")
    {
    sequence.turn_necessary();
    temperature_range.turn_necessary();
    }

  OptionMap::check_if_all_given();
  OptionMap::show_options();

  if (ran_seed.provided()) srand(ran_seed);

  Nucleotide<>::dictionary.add('U','U');
  if (duplextype.Value=="RNA")
    {
    //Nucleotide<>::dictionary.rm('T');
    Nucleotide<>::dictionary.complementary_pair('A','U');
    //std::cout << Nucleotide<>::dictionary;
    }
    
  Nucleotide<>::dictionary.add(terminal_char.Value[0],terminal_char.Value[0]);

    
  if (dict.provided())
    {
    map<string,string>::const_iterator mit;
    for(mit=dict.Value.begin(); mit!= dict.Value.end(); ++mit)
      {
      std::cout << "Adding new nucleotide to dictionary= "  << mit->first[0] << " and its complementary= " << mit->second[0] << std::endl;
      Nucleotide<>::dictionary.add(mit->first[0],mit->first[0]);
      Nucleotide<>::dictionary.add(mit->second[0],mit->second[0]);
      Nucleotide<>::dictionary.complementary_pair(mit->second[0],mit->first[0]);
      }
    }


  HeterogenousTM<value_type> HTM(expand);
  HTM.debug(debug.Value);
  HTM.verbose(debug.Value);
  HTM.matrix_base_directory(matrixbase);

  if (periodic_bc.provided()) HTM.Periodic_boundary_condition=periodic_bc;
  HTM.select_hamiltonian((std::string)model);

  //This only populates the ParameterMap, actual values are only attributed when 
  //sequences are passed to HTM
  HTM.get_parameters_from_list_of_files(par_file.Value);

  if (integration.provided()) ApplyLimit::apply(HTM,integration);
  else HTM.limits(-1.0,30.0,100);
  if (cutoff.provided()) HTM.eigenvalue_cutoff((int)cutoff);

  HTM.temperature(temperature.Value);

  ofstream outfile; outfile.open(((string)output_name+".dat").c_str());
  ofstream verfile; verfile.open(((string)output_name+".ver").c_str());
  ofstream regfile;
  ofstream texfile; texfile.open(((string)output_name+".tex").c_str());

  RegressionModel<value_type> RM(HTM);

  typedef RegressionModel<value_type>::sequence_type sequence_type;
  
  RM.Debug=debug_file;

  RM.prediction_method=prediction_method;
  
  RM.Min_length_regression=min_length_regression;
  RM.Min_a_regression=min_a_regression;

  if ((string)result=="prediction" and reg_name.provided())
    {
    cout << "Reading regression parameters from " << reg_name.Value << std::endl;
    RM.read_c(reg_name);
    }
  else regfile.open(((string)output_name+".reg").c_str());

  RM.HTM->debug(debug.Value);
  RM.HTM->verbose(debug.Value);
  
  RM.dataset.read_rules_from_list_of_files(par_file.Value);

//Read the dataset
  if(datafile.provided())
    {
    RM.dataset.read_xml((string)datafile);
    }
    
  // Here we check for the existence of a complementary strand given by -cseq
  // If -cseq was not given we try to work out the complementary strand by the rule given in Nucleotide<>::dictionary
  RegressionModel<value_type>::sequence_info_type si;
  std::string mseq, cseq;
  if (sequence.provided())
    {
    mseq=(string)sequence;
    if (csequence.provided()) cseq=(string)csequence;
    else
      {
      NucleotideSequence<> seq(sequence.Value.length()),comp(sequence.Value.length());
      seq << sequence.Value;
      comp = seq;
      comp.complementary();
      cseq = (string)comp;
      std::cout << "*** cseq not provided, worked out as " << (string)cseq <<  std::endl;
      }
    cout << "seq=" << mseq << " cseq=" << cseq << std::endl;
    
    //Individual sequences will have to be inserted into RM
    si=RM.add_temperature(mseq,cseq,salt.Value,0.0,0.0);
    HTM.pSI=&si;
    }
    
  if((string)result=="nncheck")
    {
    RM.dataset.print_temperatures(std::cout,true);
    exit(1);
    }

    
  if (sequence.provided())
    {
    if ((string)result=="averagey")
      {
      std::cout << "*** Calculating " << (string)result << " for sequence " <<  sequence.Value << std::endl;
      Looping<ActionAverageY>::print_result(HTM,0,outfile);
      }
    if ((string)result=="zyfy")
      {
      std::cout << "*** Calculating " << (string)result << " for sequence " <<  sequence.Value << std::endl;
      Looping<ActionZyFy,ApplyTemperature>::loop(HTM,temperature_range.Value,outfile,0);
      }
    if ((string)result=="prediction")
      {
      std::cout << "*** Calculating " << (string)result << " for sequence " <<  sequence.Value << std::endl;
      RM.predict(si,2);
      double temp=0.0, omega=0.0;
      temp=si.temperature.predicted;
      omega=si.thermal_equivalence;
      cout << "omega=" << omega << " Tm=" << temp << std::endl;
      RM.dataset.update(si);
      RM.get_partition_function_and_helmholtz_energy();
      RM.dataset.print_temperatures(outfile);
      }
    }

  if(datafile.provided() and ((string)result=="prediction") or ((string)result=="regression"))
    {
    std::cout << "*** Calculating " << (string)result << " for data file " <<  datafile.Value << std::endl;
    if (exp_err.provided()) RM.dataset.randomise_dataset();
    RM.recalculate_all_thermal_equivalences();
    if ((string)result=="regression") RM.calculate_regression_and_save(regfile); //otherwise reads file
    RM.predict_all();
    RM.dataset.difference();
    RM.verify(verfile);
    RM.dataset.print_temperatures(outfile);
    RM.print_tex(texfile);
    }
    
  if(datafile.provided() and ((string)result=="averagey"))
    {
    std::cout << "*** Calculating " << (string)result << " for data file " <<  datafile.Value << std::endl;
    if (exp_err.provided()) RM.dataset.randomise_dataset();
    RM.recalculate_all_thermal_equivalences();
    RM.calculate_print_all_average_y(outfile);
    }


  outfile.close();
  print_cpu();

  return 0;
}

